#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, absolute_import, print_function
from contextlib import contextmanager
from subprocess import Popen, PIPE, check_output, STDOUT
from datetime import datetime, timedelta
import xml.etree.ElementTree as ET
import sys, os, re, urllib, tempfile, shutil, codecs, socket, json, zipfile, time, pyperclip

#from pkg_resources import parse_version
from packaging import version
parse_version = version.Version

from epub_utils import epub_zip_up_book_contents, build_adobe_encryption_xml, build_idpf_encryption_xml
iswindows = sys.platform.startswith('win')
isosx = sys.platform.startswith('darwin')
islinux = sys.platform.startswith('linux')

# macOS workaround (https://www.mobileread.com/forums/showpost.php?p=3869646&postcount=241)
if isosx:
    try:
        import certifi
        os.environ["SSL_CERT_FILE"] = certifi.where()
    except:
        pass

# make sure the plugin will work with Python 2 AND Python 3
PY2 = sys.version_info[0] == 2
if PY2:
    from urllib import urlopen, urlretrieve, unquote
else:
    from urllib.request import urlopen, urlretrieve
    from urllib.parse import unquote


def jar_wrapper(*args):
    ''' wrapper for executing epubcheck.jar '''
    process = Popen(args, stdout=PIPE, stderr=PIPE, shell=False)
    ret = process.communicate()
    return ret


# code provided by DiapDealer
@contextmanager
def make_temp_directory():
    ''' creates a temporary folder '''
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)


# code provided by DiapDealer
def is_connected():
    ''' tests Internet connectivity '''
    try:
        sock = socket.create_connection(('8.8.8.8', 53), 1)
        sock.close()
        return True
    except:
        pass


def get_epc_version(epc_path):
    ''' returns the epubcheck.jar version number '''
    version = "0.0.0"

    # make sure that epubcheck.jar actually exists
    if os.path.exists(epc_path):

        # read .jar file as zip file
        archive = zipfile.ZipFile(epc_path)

        # make sure that pom.xml exists
        if 'META-INF/maven/org.w3c/epubcheck/pom.xml' in archive.namelist():
            pom_data = archive.read('META-INF/maven/org.w3c/epubcheck/pom.xml')
            archive.close()

            # parse pom.xml as ElementTree
            root = ET.fromstring(pom_data)
            tag = root.find("*//{http://maven.apache.org/POM/4.0.0}tag")
            # look for <tag>
            if tag is not None and tag.text.startswith('v'):
                version = tag.text.replace('v', '')
            else:
                # look for <version>
                project_version = root.find("{http://maven.apache.org/POM/4.0.0}version")
                if project_version is not None:
                    version = project_version.text
        else:
            print('pom.xml not found!')

    else:
        print('epubcheck.jar not found!')

    return version


def latest_epc_version(github_url):
    ''' returns the latest Github EPUBCheck version '''
    latest_version = None
    browser_download_url = ''
    if is_connected():
        try:
            response = urlopen(github_url).read().decode('utf-8')
            parsed_json = json.loads(response)
            latest_version = parsed_json[0]['tag_name']
            if latest_version.startswith('v'):
                latest_version = latest_version.replace('v', '')
                browser_download_url = parsed_json[0]['assets'][0]['browser_download_url']
            else:
                print('latest_version', latest_version)
                latest_version = None

        except Exception as ex:
            import ssl
            print('*** PYTHON ERROR ***\nSSL version:', ssl.OPENSSL_VERSION)
            exception_info = "An exception of type {0} occurred.\nArguments:\n{1!r}".format(
                type(ex).__name__, ex.args)
            print(exception_info)
            browser_download_url = '*** PYTHON ERROR *** ' + exception_info.replace('\n', ' ')
            # Wait for 2 seconds
            time.sleep(2)

    return latest_version, browser_download_url


# code provided by DiapDealer
def string_to_date(datestring):
    ''' converts dates to strings '''
    return datetime.strptime(datestring, "%Y-%m-%d %H:%M:%S.%f")


# code provided by KevinH
def generate_line_offsets(s):
    ''' returns line offsets '''
    offlst = [0]
    i = s.find('\n', 0)
    while i >= 0:
        offlst.append(i)
        i = s.find('\n', i + 1)
    # single-line files
    if s.find('\n', 0) == -1:
        offlst.append(len(s) - 1)
    return offlst


# code provided by KevinH
def charoffset(line, col, offlst):
    ''' returns character offsets  '''
    coffset = None
    if iswindows:
        coffset = offlst[line-1]  + 2 + (col - 1) - line
    else:
        coffset = offlst[line-1]  + 1 + (col - 1)
    if line == 1:
        coffset -= 1
    return coffset


# taken from navprocessor.py
def xmlencode(data):
    ''' escapes text to make it xml safe '''
    if data is None:
        return ''
    newdata = data
    newdata = newdata.replace('&', '&amp;')
    newdata = newdata.replace('<', '&lt;')
    newdata = newdata.replace('>', '&gt;')
    newdata = newdata.replace('"', '&quot;')
    return newdata


def run(bk):
    ''' main routine'''

    # get GUI language
    sigil_lang = 'en'
    locale = 'en'
    try:
        sigil_lang = bk.sigil_ui_lang[0:2]
        # check for available translations
        if sigil_lang in ['en', 'fr', 'it', 'de', 'es', 'pt']:
            locale = sigil_lang
    except:
        pass

    # get EpubCheck plugin path
    plugin_path = os.path.join(bk._w.plugin_dir, bk._w.plugin_name)

    # load translations (FIGS + Portuguese)
    import gettext
    localedir = os.path.join(plugin_path, 'locales')
    translations = gettext.translation('messages', localedir=localedir, languages=[locale])
    translations.install()
    _ = translations.gettext

    # get local lib/epubcheck.jar paths
    epc_path = os.path.join(plugin_path, 'epubcheck.jar')
    epc_lib_dir = os.path.join(plugin_path, 'lib')

    # look for Java files in plugin folder
    if os.path.exists(epc_path) and os.path.isdir(epc_lib_dir):
        epc_files_not_found = False
    else:
        epc_files_not_found = True

    # get current version number
    version = get_epc_version(epc_path)

    # get/set preference
    prefs = bk.getPrefs()

    # write initial JSON file
    if prefs == {}:
        # check if Java is installed
        if not shutil.which('java'):
            print('Java binary not found!')
            return -1
        prefs['update_check'] = True
        prefs['clipboard_copy'] = False
        prefs['usage'] = False
        prefs['github'] = True
        prefs['last_time_checked'] = str(datetime.now() - timedelta(days=14))
        prefs['check_interval'] = 14
        prefs['java_path'] = 'java'
        bk.savePrefs(prefs)

    # get preferences
    debug = prefs.get('debug', False)
    update_check = prefs.get('update_check', True)
    clipboard_copy = prefs.get('clipboard_copy', False)
    usage = prefs.get('usage', False)
    github = prefs.get('github', True)
    last_time_checked = prefs.get('last_time_checked', str(datetime.now() - timedelta(days=14)))
    check_interval = prefs.get('check_interval', 14)
    java_path = prefs.get('java_path', 'java').replace('\\\\', '/').replace('\\', '/')
    is32bit = prefs.get('is32bit', False)

    # get epubcheck language preference
    locale = None
    if 'locale' in prefs and prefs['locale'] in [
            'en',
            'de',
            'es',
            'fr',
            'it',
            'ja',
            'nl',
            'pt-BR',
            'ko-KR',
            'zh_TW',
            'da']:
        locale = prefs['locale']


    #===============================
    # run update check
    #===============================

    # get the version number of the latest EPUBCheck release
    if github or update_check:

        # make sure we have an Internet connection
        if is_connected():

            # compare current date against last update check date
            time_delta = (datetime.now() - string_to_date(last_time_checked)).days
            if (time_delta >= check_interval) or epc_files_not_found:

                #-----------------------------------
                # check for github upate
                #-----------------------------------
                if github:
                    # display update check message
                    print(_('Running update check...\n'))

                    # get latest version and download URL
                    github_url = 'https://api.github.com/repos/w3c/epubcheck/releases'
                    latest_version, browser_download_url = latest_epc_version(github_url)

                    # make sure that we have both versions, because parse_version will choke on None values
                    if (latest_version and version) and (parse_version(latest_version) > parse_version(version)) or epc_files_not_found:

                        # display New EPUBCheck version found message
                        print(_('Update found: EPUBCheck {}\n').format(latest_version))
                        base_name = os.path.basename(browser_download_url)
                        root_path = os.path.splitext(base_name)[0]

                        # create temp folder
                        with make_temp_directory() as td:
                            zip_file_name = os.path.join(td, base_name)
                            print(_('Downloading'), browser_download_url, '...')
                            urlretrieve(browser_download_url, zip_file_name)

                            # make sure the file was actually downloaded
                            if os.path.exists(zip_file_name):
                                print(browser_download_url, _('downloaded.\n'))

                                # read zip file (https://stackoverflow.com/a/19618531)
                                archive = zipfile.ZipFile(zip_file_name)
                                files = archive.namelist()
                                files_to_extract = [m for m in files \
                                if (m.startswith(root_path + '/lib') \
                                or m == root_path + '/epubcheck.jar')]
                                archive.extractall(td, files_to_extract)
                                archive.close()

                                # epubcheck.jar and /lib folder temp paths
                                temp_epc_path = os.path.join(td, root_path + '/epubcheck.jar')
                                temp_epc_lib_dir = os.path.join(td, root_path + '/lib')

                                # make sure the files were actually extracted
                                if os.path.isdir(temp_epc_lib_dir) and os.path.isfile(temp_epc_path):

                                    # delete /lib folder
                                    if os.path.isdir(epc_lib_dir):
                                        shutil.rmtree(epc_lib_dir)

                                    # delete epubcheck.jar
                                    if os.path.exists(epc_path):
                                        os.remove(epc_path)

                                    # move new files to the plugin folder
                                    shutil.move(temp_epc_lib_dir, plugin_path)
                                    shutil.move(temp_epc_path, plugin_path)

                                    # ensure you have execute rights for unix based platforms
                                    if isosx or islinux:
                                        os.chmod(epc_path, 0o744)

                                    # update version number
                                    version = latest_version

                                    # display update successful message
                                    print(_('EPUBCheck updated to {}.\n').format(version))
                                else:
                                    print(_('EPUBCheck files couldn\'t be extracted.\n'))
                            else:
                                # display Github download failed message
                                print(_('Github download failed.\n'))
                                # Wait for 2 seconds
                                time.sleep(2)

                    else:

                        if latest_version:

                            # display no update found message
                            print(_('No new EPUBCheck version found.\n'))

                        else:
                            # display update check failed message
                            print(_('Internal error: update check failed.\n'))
                            if debug:  print('Current or latest version not found.\nCurrent version: {}\nLatest version: {}\n'.format(version, latest_version))
                            # Wait for 2 seconds
                            time.sleep(2)

                            # display Python error messages;
                            # (browser_download_url contains the exception details)
                            if browser_download_url != '':
                                bk.add_result('error', None, None, browser_download_url)

                #---------------------------------------
                # check for plugin update
                #---------------------------------------
                if update_check:
                    href = 'http://www.mobileread.com/forums/showpost.php?p=2950625&postcount=1'
                    _latest_pattern = re.compile(r'Current Version:\s*&quot;([^&]*)&')
                    plugin_xml_path = os.path.abspath(os.path.join(bk._w.plugin_dir, bk._w.plugin_name, 'plugin.xml'))
                    plugin_version = ET.parse(plugin_xml_path).find('.//version').text
                    try:
                        latest_version = None
                        response = urlopen(href)
                        m = _latest_pattern.search(response.read().decode('utf-8', 'ignore'))
                        if m:
                            latest_version = (m.group(1).strip())
                            if latest_version and parse_version(latest_version) > parse_version(plugin_version):
                                restype = 'info'
                                filename = linenumber = None
                                message = _('A newer plugin version is available, version {}.').format(latest_version)
                                bk.add_result(restype, filename, linenumber, message)
                                if debug: print(message)

                    except:
                        print(_('Internal error: latest plugin version not found.\n'))

            # update time stamp
            prefs['last_time_checked'] = str(datetime.now())
            bk.savePrefs(prefs)

        else:
            # display update skipped message
            print(_('Update check skipped: no Internet.\n'))
            # Wait for 2 seconds
            time.sleep(2)

    #======================
    # run epubcheck
    #======================
    if not os.path.exists(epc_path) or not os.path.isdir(epc_lib_dir):
        print(_('epubcheck.java files not found! Please re-install the plugin.'))
        return -1

    #-------------------------------------------------
    # get opf file name and contents
    #-------------------------------------------------

    # get the default opf path and the opf file name
    if bk.launcher_version() >= 20190927:
        opf_path = os.path.join(bk._w.ebook_root, bk.get_opfbookpath())
        opf_name = os.path.basename(opf_path)
    else:
        opf_name = bk._w.opfname
        opf_path = os.path.join(bk._w.ebook_root, 'OEBPS', opf_name)

    # read the original opf file
    if os.path.isfile(opf_path):
        if PY2:
            f = codecs.open(opf_path, 'r', encoding='utf8')
        else:
            f = open(opf_path, 'r', encoding='utf-8')
        opf_contents = f.read()
        f.close()

    else:
        print(_('Internal error: opf path not found!'), opf_path)
        return -1

    #-------------------------------------------------------------
    # copy current epub to temp epub
    #-------------------------------------------------------------
    with make_temp_directory() as temp_dir:

        # copy book files
        bk.copy_book_contents_to(temp_dir)

        #=======================================================
        # handle encrption.xml
        #=======================================================
        
        # create encryption.xml from scratch  (requires Sigil 2.2.0 or higher)
        if bk.launcher_version() >= 20240522:  
            meta_inf_path = os.path.join(temp_dir, 'META-INF', 'encryption.xml')
            mangled_fonts = []
            preferred_algorithm = None
            encryption_xml_lines = None
            for mid, href, mime in bk.font_iter():
                # get obfuscated fonts and algorithms
                bookhref = bk.id_to_bookpath(mid)
                preferred_algorithm = bk.font_bookpath_to_preferred_obfuscation_algorithm(bookhref)
                if preferred_algorithm: mangled_fonts.append(bookhref)
            
            # generate encryption.xml
            if mangled_fonts != []:
                # use first algorithm as default, in case only some of the fonts were obfuscated
                preferred_algorithm = bk.font_bookpath_to_preferred_obfuscation_algorithm(mangled_fonts[0])
                encryption_xml = '<?xml version="1.0" encoding="UTF-8"?>\n'
                if preferred_algorithm == 'http://ns.adobe.com/pdf/enc#RC':
                    encryption_xml += build_adobe_encryption_xml(mangled_fonts)
                else:
                    encryption_xml += build_idpf_encryption_xml(mangled_fonts)
                if debug: print('encryption.xml\n\n', encryption_xml)
                encryption_xml_lines = encryption_xml.splitlines()
                # save encryption.xml in temp epub folder
                with open(meta_inf_path, 'w', encoding='utf-8') as fh:
                    fh.write(encryption_xml)
                    if debug: bk.add_result('info', None, None, 'encryption.xml created from scratch')
            
        # extract encryption.xml from original epub file ( requires Sigil 0.9.10 or higher)
        elif bk.launcher_version() >= 20180122:  
            filepath = bk.get_epub_filepath()
            if filepath != "":
                encryption_xml_lines = None
                orig_epub = zipfile.ZipFile(filepath)
                if 'META-INF/encryption.xml' in orig_epub.namelist():
                    encryption_xml = orig_epub.read('META-INF/encryption.xml').decode('UTF-8')
                    if debug: print('encryption.xml\n\n', encryption_xml)
                    encryption_xml_lines = encryption_xml.splitlines()
                    orig_epub.extract('META-INF/encryption.xml', temp_dir)
                    if debug: bk.add_result('info', None, None, 'encryption.xml copied from original epub')

        else:
            # encrpytion.xml can't be extracted from the epub in older versions
            encryption_xml_lines = None
            if debug: bk.add_result('info', None, None, 'encryption.xml not supported')

        # create mimetype file
        with open(os.path.join(temp_dir, "mimetype"), "w") as mimetype:
            mimetype.write("application/epub+zip")

        # zip up the epub folder and save the epub in the plugin folder
        epub_path = os.path.join(bk._w.plugin_dir, bk._w.plugin_name, 'temp.epub')
        if os.path.isfile(epub_path):
            os.remove(str(epub_path))
        epub_zip_up_book_contents(temp_dir, epub_path)

    #----------------------------------------------------------------------
    # define epubcheck command line parameters
    #----------------------------------------------------------------------
    if is32bit:
        args = [java_path, '-Dfile.encoding=UTF8', '-Xss1024k', '-jar', epc_path, epub_path, '-q', '--json', '-']
    else:
        args = [java_path, '-Dfile.encoding=UTF8', '-jar', epc_path, epub_path, '-q', '--json', '-']

    # change message language
    if locale:
        args.extend(['--locale', locale])

    # display usage messages
    if usage:
        args.append('--usage')

    # display status message
    print(_('Running EPUBCheck {}... please wait.').format(version))


    #--------------------------------
    # run epubcheck
    #--------------------------------
    if debug: 
        print('\nargs:', *args)

    try:
        result = jar_wrapper(*args)
        stdout = result[0].decode("utf-8")
        stderr = result[1].decode("utf-8")

    except Exception as ex:
        if type(ex).__name__ == 'FileNotFoundError':
            print('The Java Runtime Environment couldn\'t be found!!!')
        else:
            print('An exception of type {0} occurred.'.format(type(ex).__name__))
        return -1

    if debug: 
        print(stdout, stderr)

    # check for StackOverflowError/Exception errors
    if stderr.find('java.lang.StackOverflowError') != -1 or stderr.find('Exception in thread') != -1:
        print(_('EPUBCheck Java error.\n'), stderr)
        return -1

    # delete temp epub from plugin folder
    os.remove(str(epub_path))

    # copy error messages to clipboard
    if clipboard_copy:
        pyperclip.copy(stderr+stdout)

    # reset file offset dictionary
    offsets = {}

    #---------------------------------
    # parse JSON file
    #---------------------------------

    parsed_json = json.loads(stdout)
    if 'messages' in parsed_json:
        messages = parsed_json['messages']
        for message in messages:

            # get severity
            severity = message['severity']

            # map severity to Sigil restypes
            if severity in ['ERROR', 'FATAL']:
                restype = 'error'
            elif severity == 'WARNING':
                restype = 'warning'
            else:
                restype = 'info'

            # get remaining items
            ID = message['ID']
            msg = message['message']
            locations = message['locations']
            number_of_locations = len(locations)

            # parse messages
            for location in range(number_of_locations):
                # get error info
                path = locations[location]['path']

                # sometimes EPUBCheck returns the epub path instead of the file path 
                if path.endswith('.epub'):
                    # try to get the file path from the error message
                    msg_path_re = re.search('\"([^\"]+)\"', msg)
                    if len(msg_path_re.groups()) == 1:
                        path =  msg_path_re.group(1)

                # check for file names with escaped spaces %20%
                if '%' in path: path = unquote(path) 

                # get file name and manifest id
                filename = os.path.basename(path)
                manifest_id = bk.basename_to_id(filename)

                # older Sigil versions don't require the full file path
                if bk.launcher_version() < 20190927:
                    path = filename

                line = locations[location]['line']
                column = locations[location]['column']
                coffset = None

                # get or generate line offsets
                if line != -1 and column != -1 and filename != 'encryption.xml' and bk.launcher_version() >= 20160909:

                    # calculate line offset
                    if manifest_id not in offsets:

                        # get the text
                        text = None
                        if filename == opf_name:
                            text = opf_contents
                        else:
                            try:
                                text = bk.readfile(bk.basename_to_id(filename))
                            except:
                                if debug: bk.add_result('error', path, None, _('*** Internal error. bk.readfile() failed for: ') + filename)

                        # calculate line offsets
                        if text and not 'DOCTYPE svg' in text:
                            offlst = generate_line_offsets(text)
                            offsets[manifest_id] = offlst
                        else:
                            offlst = [0]

                    else:
                        offlst = offsets[manifest_id]

                    # calculate character offset
                    if offlst != [0]:
                        coffset = charoffset(int(line), int(column), offlst)
                        if iswindows and filename == opf_name:
                            coffset += int(line) - 1

                # try to get font  file name from encryption.xml
                if path == 'META-INF/encryption.xml':   
                    if encryption_xml_lines and len(encryption_xml_lines)  >=  int(line) - 1:
                        enc_line = encryption_xml_lines[int(line) - 1]
                        # <enc:CipherReference URI="OEBPS/Fonts/NotoEmoji-Bold.ttf"/>
                        enc_line_parts = enc_line.split('"')
                        if len(enc_line_parts)  == 3:
                            path = enc_line_parts[1]
                    line = column = -1

                # format message
                xmsg = xmlencode('Col: {}: {}({}): {}'.format(column, severity, ID, msg))

                # add message to validation pane
                if coffset:
                    bk.add_extended_result(restype, path, line, coffset, xmsg)
                else:
                    bk.add_result(restype, path, line, xmsg)

        if debug:
            # display miscellaneous debugging information
            checker = parsed_json['checker']
            publication = parsed_json['publication']
            epub_version = str(publication['ePubVersion'])
            for key in checker:
                if key not in ['path', 'filename']:
                    bk.add_result('info', None, None, key + ': ' +  str(checker[key]))

            if epub_version.startswith('3'):
                # according to epc, an epub3 is compatible, if:
                # the nav is NOT in the spine
                # none of the html files contains HTML5 tags
                # an NCX file was found
                bk.add_result('info', None, None, 'isBackwardCompatible: ' + str(publication['isBackwardCompatible']))

                # display file information
                items = parsed_json['items']
                for item in items:
                    if item['media_type'] == 'application/xhtml+xml':
                        path = item['fileName']
                        for key in item:
                            item_key = item[key]
                            if item_key == True or key == 'spineIndex':
                                bk.add_result('info', path, None, key + ': ' + str(item[key]))

    return 0

def main():
    print('I reached main when I should not have\n')
    return -1

if __name__ == "__main__":
    sys.exit(main())
